@@ -58,6 +58,7 @@
 #include <string.h>
 
 #define VKD3D_VEC4_SIZE 4
+#define VKD3D_DVEC2_SIZE 2
 
 enum vkd3d_shader_error
 {
@@ -69,19 +70,70 @@ enum vkd3d_shader_error
     VKD3D_SHADER_ERROR_DXBC_INVALID_CHUNK_SIZE          = 6,
 
     VKD3D_SHADER_ERROR_TPF_MISMATCHED_CF                = 1000,
+    VKD3D_SHADER_ERROR_TPF_INVALID_REGISTER_RANGE       = 1001,
 
     VKD3D_SHADER_ERROR_SPV_DESCRIPTOR_BINDING_NOT_FOUND = 2000,
     VKD3D_SHADER_ERROR_SPV_INVALID_REGISTER_TYPE        = 2001,
     VKD3D_SHADER_ERROR_SPV_INVALID_DESCRIPTOR_BINDING   = 2002,
+    VKD3D_SHADER_ERROR_SPV_DESCRIPTOR_IDX_UNSUPPORTED   = 2003,
+    VKD3D_SHADER_ERROR_SPV_STENCIL_EXPORT_UNSUPPORTED   = 2004,
 
     VKD3D_SHADER_ERROR_RS_OUT_OF_MEMORY                 = 3000,
     VKD3D_SHADER_ERROR_RS_INVALID_VERSION               = 3001,
     VKD3D_SHADER_ERROR_RS_INVALID_ROOT_PARAMETER_TYPE   = 3002,
     VKD3D_SHADER_ERROR_RS_INVALID_DESCRIPTOR_RANGE_TYPE = 3003,
     VKD3D_SHADER_ERROR_RS_MIXED_DESCRIPTOR_RANGE_TYPES  = 3004,
+
+    VKD3D_SHADER_ERROR_PP_INVALID_SYNTAX                = 4000,
+    VKD3D_SHADER_ERROR_PP_ERROR_DIRECTIVE               = 4001,
+    VKD3D_SHADER_ERROR_PP_INCLUDE_FAILED                = 4002,
+
+    VKD3D_SHADER_WARNING_PP_ALREADY_DEFINED             = 4300,
+    VKD3D_SHADER_WARNING_PP_INVALID_DIRECTIVE           = 4301,
+    VKD3D_SHADER_WARNING_PP_ARGUMENT_COUNT_MISMATCH     = 4302,
+    VKD3D_SHADER_WARNING_PP_UNKNOWN_DIRECTIVE           = 4303,
+    VKD3D_SHADER_WARNING_PP_UNTERMINATED_MACRO          = 4304,
+    VKD3D_SHADER_WARNING_PP_UNTERMINATED_IF             = 4305,
+    VKD3D_SHADER_WARNING_PP_DIV_BY_ZERO                 = 4306,
+
+    VKD3D_SHADER_ERROR_HLSL_INVALID_SYNTAX              = 5000,
+    VKD3D_SHADER_ERROR_HLSL_INVALID_MODIFIER            = 5001,
+    VKD3D_SHADER_ERROR_HLSL_INVALID_TYPE                = 5002,
+    VKD3D_SHADER_ERROR_HLSL_MODIFIES_CONST              = 5003,
+    VKD3D_SHADER_ERROR_HLSL_MISSING_SEMANTIC            = 5004,
+    VKD3D_SHADER_ERROR_HLSL_NOT_DEFINED                 = 5005,
+    VKD3D_SHADER_ERROR_HLSL_REDEFINED                   = 5006,
+    VKD3D_SHADER_ERROR_HLSL_WRONG_PARAMETER_COUNT       = 5007,
+    VKD3D_SHADER_ERROR_HLSL_INVALID_SIZE                = 5008,
+    VKD3D_SHADER_ERROR_HLSL_MISSING_INITIALIZER         = 5009,
+    VKD3D_SHADER_ERROR_HLSL_INVALID_LVALUE              = 5010,
+    VKD3D_SHADER_ERROR_HLSL_INVALID_WRITEMASK           = 5011,
+    VKD3D_SHADER_ERROR_HLSL_INVALID_INDEX               = 5012,
+    VKD3D_SHADER_ERROR_HLSL_INVALID_SEMANTIC            = 5013,
+    VKD3D_SHADER_ERROR_HLSL_INVALID_RETURN              = 5014,
+    VKD3D_SHADER_ERROR_HLSL_OVERLAPPING_RESERVATIONS    = 5015,
+    VKD3D_SHADER_ERROR_HLSL_INVALID_RESERVATION         = 5016,
+    VKD3D_SHADER_ERROR_HLSL_NOT_IMPLEMENTED             = 5017,
+    VKD3D_SHADER_ERROR_HLSL_INVALID_TEXEL_OFFSET        = 5018,
+    VKD3D_SHADER_ERROR_HLSL_OFFSET_OUT_OF_BOUNDS        = 5019,
+    VKD3D_SHADER_ERROR_HLSL_INCOMPATIBLE_PROFILE        = 5020,
+    VKD3D_SHADER_ERROR_HLSL_DIVISION_BY_ZERO            = 5021,
+    VKD3D_SHADER_ERROR_HLSL_NON_STATIC_OBJECT_REF       = 5022,
+
+    VKD3D_SHADER_WARNING_HLSL_IMPLICIT_TRUNCATION       = 5300,
+    VKD3D_SHADER_WARNING_HLSL_DIVISION_BY_ZERO          = 5301,
+
+    VKD3D_SHADER_ERROR_GLSL_INTERNAL                    = 6000,
+
+    VKD3D_SHADER_ERROR_D3DBC_UNEXPECTED_EOF             = 7000,
+    VKD3D_SHADER_ERROR_D3DBC_INVALID_VERSION_TOKEN      = 7001,
+    VKD3D_SHADER_ERROR_D3DBC_INVALID_OPCODE             = 7002,
+    VKD3D_SHADER_ERROR_D3DBC_INVALID_RESOURCE_TYPE      = 7003,
+
+    VKD3D_SHADER_WARNING_D3DBC_IGNORED_INSTRUCTION_FLAGS= 7300,
 };
 
-enum VKD3D_SHADER_INSTRUCTION_HANDLER
+enum vkd3d_shader_opcode
 {
     VKD3DSIH_ABS,
     VKD3DSIH_ADD,
@@ -105,6 +157,7 @@ enum VKD3D_SHADER_INSTRUCTION_HANDLER
     VKD3DSIH_CALL,
     VKD3DSIH_CALLNZ,
     VKD3DSIH_CASE,
+    VKD3DSIH_CHECK_ACCESS_FULLY_MAPPED,
     VKD3DSIH_CMP,
     VKD3DSIH_CND,
     VKD3DSIH_CONTINUE,
@@ -113,6 +166,7 @@ enum VKD3D_SHADER_INSTRUCTION_HANDLER
     VKD3DSIH_CRS,
     VKD3DSIH_CUT,
     VKD3DSIH_CUT_STREAM,
+    VKD3DSIH_DADD,
     VKD3DSIH_DCL,
     VKD3DSIH_DCL_CONSTANT_BUFFER,
     VKD3DSIH_DCL_FUNCTION_BODY,
@@ -153,15 +207,27 @@ enum VKD3D_SHADER_INSTRUCTION_HANDLER
     VKD3DSIH_DCL_UAV_STRUCTURED,
     VKD3DSIH_DCL_UAV_TYPED,
     VKD3DSIH_DCL_VERTICES_OUT,
+    VKD3DSIH_DDIV,
     VKD3DSIH_DEF,
     VKD3DSIH_DEFAULT,
     VKD3DSIH_DEFB,
     VKD3DSIH_DEFI,
+    VKD3DSIH_DEQ,
+    VKD3DSIH_DFMA,
+    VKD3DSIH_DGE,
     VKD3DSIH_DIV,
+    VKD3DSIH_DLT,
+    VKD3DSIH_DMAX,
+    VKD3DSIH_DMIN,
+    VKD3DSIH_DMOV,
+    VKD3DSIH_DMOVC,
+    VKD3DSIH_DMUL,
+    VKD3DSIH_DNE,
     VKD3DSIH_DP2,
     VKD3DSIH_DP2ADD,
     VKD3DSIH_DP3,
     VKD3DSIH_DP4,
+    VKD3DSIH_DRCP,
     VKD3DSIH_DST,
     VKD3DSIH_DSX,
     VKD3DSIH_DSX_COARSE,
@@ -169,6 +235,9 @@ enum VKD3D_SHADER_INSTRUCTION_HANDLER
     VKD3DSIH_DSY,
     VKD3DSIH_DSY_COARSE,
     VKD3DSIH_DSY_FINE,
+    VKD3DSIH_DTOF,
+    VKD3DSIH_DTOI,
+    VKD3DSIH_DTOU,
     VKD3DSIH_ELSE,
     VKD3DSIH_EMIT,
     VKD3DSIH_EMIT_STREAM,
@@ -188,12 +257,17 @@ enum VKD3D_SHADER_INSTRUCTION_HANDLER
     VKD3DSIH_FIRSTBIT_LO,
     VKD3DSIH_FIRSTBIT_SHI,
     VKD3DSIH_FRC,
+    VKD3DSIH_FTOD,
     VKD3DSIH_FTOI,
     VKD3DSIH_FTOU,
     VKD3DSIH_GATHER4,
     VKD3DSIH_GATHER4_C,
+    VKD3DSIH_GATHER4_C_S,
     VKD3DSIH_GATHER4_PO,
     VKD3DSIH_GATHER4_PO_C,
+    VKD3DSIH_GATHER4_PO_C_S,
+    VKD3DSIH_GATHER4_PO_S,
+    VKD3DSIH_GATHER4_S,
     VKD3DSIH_GE,
     VKD3DSIH_HS_CONTROL_POINT_PHASE,
     VKD3DSIH_HS_DECLS,
@@ -226,13 +300,19 @@ enum VKD3D_SHADER_INSTRUCTION_HANDLER
     VKD3DSIH_INEG,
     VKD3DSIH_ISHL,
     VKD3DSIH_ISHR,
+    VKD3DSIH_ITOD,
     VKD3DSIH_ITOF,
     VKD3DSIH_LABEL,
     VKD3DSIH_LD,
     VKD3DSIH_LD2DMS,
+    VKD3DSIH_LD2DMS_S,
     VKD3DSIH_LD_RAW,
+    VKD3DSIH_LD_RAW_S,
+    VKD3DSIH_LD_S,
     VKD3DSIH_LD_STRUCTURED,
+    VKD3DSIH_LD_STRUCTURED_S,
     VKD3DSIH_LD_UAV_TYPED,
+    VKD3DSIH_LD_UAV_TYPED_S,
     VKD3DSIH_LIT,
     VKD3DSIH_LOD,
     VKD3DSIH_LOG,
@@ -251,6 +331,7 @@ enum VKD3D_SHADER_INSTRUCTION_HANDLER
     VKD3DSIH_MOV,
     VKD3DSIH_MOVA,
     VKD3DSIH_MOVC,
+    VKD3DSIH_MSAD,
     VKD3DSIH_MUL,
     VKD3DSIH_NE,
     VKD3DSIH_NOP,
@@ -271,11 +352,17 @@ enum VKD3D_SHADER_INSTRUCTION_HANDLER
     VKD3DSIH_RSQ,
     VKD3DSIH_SAMPLE,
     VKD3DSIH_SAMPLE_B,
+    VKD3DSIH_SAMPLE_B_CL_S,
     VKD3DSIH_SAMPLE_C,
+    VKD3DSIH_SAMPLE_C_CL_S,
     VKD3DSIH_SAMPLE_C_LZ,
+    VKD3DSIH_SAMPLE_C_LZ_S,
+    VKD3DSIH_SAMPLE_CL_S,
     VKD3DSIH_SAMPLE_GRAD,
+    VKD3DSIH_SAMPLE_GRAD_CL_S,
     VKD3DSIH_SAMPLE_INFO,
     VKD3DSIH_SAMPLE_LOD,
+    VKD3DSIH_SAMPLE_LOD_S,
     VKD3DSIH_SAMPLE_POS,
     VKD3DSIH_SETP,
     VKD3DSIH_SGE,
@@ -320,6 +407,7 @@ enum VKD3D_SHADER_INSTRUCTION_HANDLER
     VKD3DSIH_UMIN,
     VKD3DSIH_UMUL,
     VKD3DSIH_USHR,
+    VKD3DSIH_UTOD,
     VKD3DSIH_UTOF,
     VKD3DSIH_XOR,
 
@@ -351,6 +439,7 @@ enum vkd3d_shader_register_type
     VKD3DSPR_LABEL = 18,
     VKD3DSPR_PREDICATE = 19,
     VKD3DSPR_IMMCONST,
+    VKD3DSPR_IMMCONST64,
     VKD3DSPR_CONSTBUFFER,
     VKD3DSPR_IMMCONSTBUFFER,
     VKD3DSPR_PRIMID,
@@ -379,10 +468,22 @@ enum vkd3d_shader_register_type
     VKD3DSPR_DEPTHOUTGE,
     VKD3DSPR_DEPTHOUTLE,
     VKD3DSPR_RASTERIZER,
+    VKD3DSPR_OUTSTENCILREF,
 
     VKD3DSPR_INVALID = ~0u,
 };
 
+enum vkd3d_shader_register_precision
+{
+    VKD3D_SHADER_REGISTER_PRECISION_DEFAULT,
+    VKD3D_SHADER_REGISTER_PRECISION_MIN_FLOAT_16,
+    VKD3D_SHADER_REGISTER_PRECISION_MIN_FLOAT_10,
+    VKD3D_SHADER_REGISTER_PRECISION_MIN_INT_16,
+    VKD3D_SHADER_REGISTER_PRECISION_MIN_UINT_16,
+
+    VKD3D_SHADER_REGISTER_PRECISION_INVALID = ~0u,
+};
+
 enum vkd3d_data_type
 {
     VKD3D_DATA_FLOAT,
@@ -394,6 +495,10 @@ enum vkd3d_data_type
     VKD3D_DATA_UNORM,
     VKD3D_DATA_SNORM,
     VKD3D_DATA_OPAQUE,
+    VKD3D_DATA_MIXED,
+    VKD3D_DATA_DOUBLE,
+    VKD3D_DATA_CONTINUED,
+    VKD3D_DATA_UNUSED,
 };
 
 enum vkd3d_immconst_type
@@ -449,16 +554,19 @@ enum vkd3d_shader_interpolation_mode
 enum vkd3d_shader_global_flags
 {
     VKD3DSGF_REFACTORING_ALLOWED               = 0x01,
+    VKD3DSGF_ENABLE_DOUBLE_PRECISION_FLOAT_OPS = 0x02,
     VKD3DSGF_FORCE_EARLY_DEPTH_STENCIL         = 0x04,
     VKD3DSGF_ENABLE_RAW_AND_STRUCTURED_BUFFERS = 0x08,
     VKD3DSGF_SKIP_OPTIMIZATION                 = 0x10,
-    VKD3DSGF_ENABLE_MINIMUM_PRECISION          = 0x20
+    VKD3DSGF_ENABLE_MINIMUM_PRECISION          = 0x20,
+    VKD3DSGF_ENABLE_11_1_DOUBLE_EXTENSIONS     = 0x40,
 };
 
 enum vkd3d_shader_sync_flags
 {
     VKD3DSSF_THREAD_GROUP        = 0x1,
     VKD3DSSF_GROUP_SHARED_MEMORY = 0x2,
+    VKD3DSSF_GLOBAL_UAV          = 0x8,
 };
 
 enum vkd3d_shader_uav_flags
@@ -506,19 +614,6 @@ enum vkd3d_shader_conditional_op
     VKD3D_SHADER_CONDITIONAL_OP_Z  = 1
 };
 
-#define VKD3D_SM1_VS  0xfffeu
-#define VKD3D_SM1_PS  0xffffu
-#define VKD3D_SM4_PS  0x0000u
-#define VKD3D_SM4_VS  0x0001u
-#define VKD3D_SM4_GS  0x0002u
-#define VKD3D_SM5_HS  0x0003u
-#define VKD3D_SM5_DS  0x0004u
-#define VKD3D_SM5_CS  0x0005u
-
-/* Shader version tokens, and shader end tokens */
-#define VKD3DPS_VERSION(major, minor) ((VKD3D_SM1_PS << 16) | ((major) << 8) | (minor))
-#define VKD3DVS_VERSION(major, minor) ((VKD3D_SM1_VS << 16) | ((major) << 8) | (minor))
-
 #define MAX_IMMEDIATE_CONSTANT_BUFFER_SIZE 4096
 #define MAX_REG_OUTPUT 32
 
@@ -532,6 +627,10 @@ enum vkd3d_shader_type
     VKD3D_SHADER_TYPE_GRAPHICS_COUNT,
 
     VKD3D_SHADER_TYPE_COMPUTE = VKD3D_SHADER_TYPE_GRAPHICS_COUNT,
+
+    VKD3D_SHADER_TYPE_EFFECT,
+    VKD3D_SHADER_TYPE_TEXTURE,
+    VKD3D_SHADER_TYPE_LIBRARY,
     VKD3D_SHADER_TYPE_COUNT,
 };
 
@@ -545,7 +644,7 @@ struct vkd3d_shader_version
 struct vkd3d_shader_immediate_constant_buffer
 {
     unsigned int vec4_count;
-    DWORD data[MAX_IMMEDIATE_CONSTANT_BUFFER_SIZE];
+    uint32_t data[MAX_IMMEDIATE_CONSTANT_BUFFER_SIZE];
 };
 
 struct vkd3d_shader_indexable_temp
@@ -565,6 +664,8 @@ struct vkd3d_shader_register_index
 struct vkd3d_shader_register
 {
     enum vkd3d_shader_register_type type;
+    enum vkd3d_shader_register_precision precision;
+    bool non_uniform;
     enum vkd3d_data_type data_type;
     struct vkd3d_shader_register_index idx[3];
     enum vkd3d_immconst_type immconst_type;
@@ -572,6 +673,8 @@ struct vkd3d_shader_register
     {
         DWORD immconst_uint[VKD3D_VEC4_SIZE];
         float immconst_float[VKD3D_VEC4_SIZE];
+        uint64_t immconst_uint64[VKD3D_DVEC2_SIZE];
+        double immconst_double[VKD3D_DVEC2_SIZE];
         unsigned fp_body_idx;
     } u;
 };
@@ -597,11 +700,16 @@ struct vkd3d_shader_index_range
     unsigned int register_count;
 };
 
+struct vkd3d_shader_register_range
+{
+    unsigned int space;
+    unsigned int first, last;
+};
+
 struct vkd3d_shader_resource
 {
     struct vkd3d_shader_dst_param reg;
-    unsigned int register_space;
-    unsigned int register_index;
+    struct vkd3d_shader_register_range range;
 };
 
 enum vkd3d_decl_usage
@@ -627,7 +735,7 @@ struct vkd3d_shader_semantic
     enum vkd3d_decl_usage usage;
     unsigned int usage_idx;
     enum vkd3d_shader_resource_type resource_type;
-    enum vkd3d_data_type resource_data_type;
+    enum vkd3d_data_type resource_data_type[VKD3D_VEC4_SIZE];
     struct vkd3d_shader_resource resource;
 };
 
@@ -660,7 +768,7 @@ enum vkd3d_shader_input_sysval_semantic
 
 struct vkd3d_shader_desc
 {
-    const DWORD *byte_code;
+    const uint32_t *byte_code;
     size_t byte_code_size;
     struct vkd3d_shader_signature input_signature;
     struct vkd3d_shader_signature output_signature;
@@ -676,14 +784,14 @@ struct vkd3d_shader_register_semantic
 struct vkd3d_shader_sampler
 {
     struct vkd3d_shader_src_param src;
-    unsigned int register_space, register_index;
+    struct vkd3d_shader_register_range range;
 };
 
 struct vkd3d_shader_constant_buffer
 {
     struct vkd3d_shader_src_param src;
     unsigned int size;
-    unsigned int register_space, register_index;
+    struct vkd3d_shader_register_range range;
 };
 
 struct vkd3d_shader_structured_resource
@@ -758,7 +866,7 @@ struct vkd3d_shader_primitive_type
 
 struct vkd3d_shader_instruction
 {
-    enum VKD3D_SHADER_INSTRUCTION_HANDLER handler_idx;
+    enum vkd3d_shader_opcode handler_idx;
     DWORD flags;
     unsigned int dst_count;
     unsigned int src_count;
@@ -766,8 +874,9 @@ struct vkd3d_shader_instruction
     const struct vkd3d_shader_src_param *src;
     struct vkd3d_shader_texel_offset texel_offset;
     enum vkd3d_shader_resource_type resource_type;
-    enum vkd3d_data_type resource_data_type;
-    bool coissue;
+    unsigned int resource_stride;
+    enum vkd3d_data_type resource_data_type[VKD3D_VEC4_SIZE];
+    bool coissue, structured, raw;
     const struct vkd3d_shader_src_param *predicate;
     union
     {
@@ -810,71 +919,191 @@ static inline bool vkd3d_shader_register_is_output(const struct vkd3d_shader_reg
     return reg->type == VKD3DSPR_OUTPUT || reg->type == VKD3DSPR_COLOROUT;
 }
 
-void vkd3d_shader_trace(void *data) DECLSPEC_HIDDEN;
+struct vkd3d_shader_location
+{
+    const char *source_name;
+    unsigned int line, column;
+};
 
-const char *shader_get_type_prefix(enum vkd3d_shader_type type) DECLSPEC_HIDDEN;
+struct vkd3d_shader_parser
+{
+    struct vkd3d_shader_message_context *message_context;
+    struct vkd3d_shader_location location;
+    bool failed;
 
-void *shader_sm4_init(const DWORD *byte_code, size_t byte_code_size,
-        const struct vkd3d_shader_signature *output_signature) DECLSPEC_HIDDEN;
-void shader_sm4_free(void *data) DECLSPEC_HIDDEN;
-void shader_sm4_read_header(void *data, const DWORD **ptr,
-        struct vkd3d_shader_version *shader_version) DECLSPEC_HIDDEN;
-void shader_sm4_read_instruction(void *data, const DWORD **ptr,
-        struct vkd3d_shader_instruction *ins) DECLSPEC_HIDDEN;
-bool shader_sm4_is_end(void *data, const DWORD **ptr) DECLSPEC_HIDDEN;
+    struct vkd3d_shader_desc shader_desc;
+    struct vkd3d_shader_version shader_version;
+    const uint32_t *ptr;
+    const struct vkd3d_shader_parser_ops *ops;
+};
+
+struct vkd3d_shader_parser_ops
+{
+    void (*parser_reset)(struct vkd3d_shader_parser *parser);
+    void (*parser_destroy)(struct vkd3d_shader_parser *parser);
+    void (*parser_read_instruction)(struct vkd3d_shader_parser *parser, struct vkd3d_shader_instruction *instruction);
+    bool (*parser_is_end)(struct vkd3d_shader_parser *parser);
+};
+
+void vkd3d_shader_parser_error(struct vkd3d_shader_parser *parser,
+        enum vkd3d_shader_error error, const char *format, ...) VKD3D_PRINTF_FUNC(3, 4);
+void vkd3d_shader_parser_init(struct vkd3d_shader_parser *parser,
+        struct vkd3d_shader_message_context *message_context, const char *source_name,
+        const struct vkd3d_shader_version *version, const struct vkd3d_shader_parser_ops *ops);
+void vkd3d_shader_parser_warning(struct vkd3d_shader_parser *parser,
+        enum vkd3d_shader_error error, const char *format, ...) VKD3D_PRINTF_FUNC(3, 4);
+
+static inline void vkd3d_shader_parser_destroy(struct vkd3d_shader_parser *parser)
+{
+    parser->ops->parser_destroy(parser);
+}
+
+static inline bool vkd3d_shader_parser_is_end(struct vkd3d_shader_parser *parser)
+{
+    return parser->ops->parser_is_end(parser);
+}
+
+static inline void vkd3d_shader_parser_read_instruction(struct vkd3d_shader_parser *parser,
+        struct vkd3d_shader_instruction *instruction)
+{
+    parser->ops->parser_read_instruction(parser, instruction);
+}
+
+static inline void vkd3d_shader_parser_reset(struct vkd3d_shader_parser *parser)
+{
+    parser->ops->parser_reset(parser);
+}
+
+void vkd3d_shader_trace(struct vkd3d_shader_parser *parser);
+
+const char *shader_get_type_prefix(enum vkd3d_shader_type type);
 
 struct vkd3d_string_buffer
 {
     char *buffer;
-    unsigned int buffer_size;
-    unsigned int content_size;
+    size_t buffer_size, content_size;
 };
 
-void vkd3d_string_buffer_cleanup(struct vkd3d_string_buffer *buffer) DECLSPEC_HIDDEN;
-bool vkd3d_string_buffer_init(struct vkd3d_string_buffer *buffer) DECLSPEC_HIDDEN;
-int vkd3d_string_buffer_vprintf(struct vkd3d_string_buffer *buffer, const char *format, va_list args) DECLSPEC_HIDDEN;
+struct vkd3d_string_buffer_cache
+{
+    struct vkd3d_string_buffer **buffers;
+    size_t count, max_count, capacity;
+};
+
+enum vkd3d_result vkd3d_dxbc_binary_to_text(struct vkd3d_shader_parser *parser,
+        const struct vkd3d_shader_compile_info *compile_info, struct vkd3d_shader_code *out);
+void vkd3d_string_buffer_cleanup(struct vkd3d_string_buffer *buffer);
+struct vkd3d_string_buffer *vkd3d_string_buffer_get(struct vkd3d_string_buffer_cache *list);
+void vkd3d_string_buffer_init(struct vkd3d_string_buffer *buffer);
+void vkd3d_string_buffer_cache_cleanup(struct vkd3d_string_buffer_cache *list);
+void vkd3d_string_buffer_cache_init(struct vkd3d_string_buffer_cache *list);
+int vkd3d_string_buffer_print_f32(struct vkd3d_string_buffer *buffer, float f);
+int vkd3d_string_buffer_print_f64(struct vkd3d_string_buffer *buffer, double d);
+int vkd3d_string_buffer_printf(struct vkd3d_string_buffer *buffer, const char *format, ...) VKD3D_PRINTF_FUNC(2, 3);
+void vkd3d_string_buffer_release(struct vkd3d_string_buffer_cache *list, struct vkd3d_string_buffer *buffer);
+#define vkd3d_string_buffer_trace(buffer) \
+        vkd3d_string_buffer_trace_(buffer, __FUNCTION__)
+void vkd3d_string_buffer_trace_(const struct vkd3d_string_buffer *buffer, const char *function);
+int vkd3d_string_buffer_vprintf(struct vkd3d_string_buffer *buffer, const char *format, va_list args);
+
+struct vkd3d_bytecode_buffer
+{
+    uint8_t *data;
+    size_t size, capacity;
+    int status;
+};
+
+size_t bytecode_put_bytes(struct vkd3d_bytecode_buffer *buffer, const void *bytes, size_t size);
+void set_u32(struct vkd3d_bytecode_buffer *buffer, size_t offset, uint32_t value);
+
+static inline size_t put_u32(struct vkd3d_bytecode_buffer *buffer, uint32_t value)
+{
+    return bytecode_put_bytes(buffer, &value, sizeof(value));
+}
+
+static inline size_t put_f32(struct vkd3d_bytecode_buffer *buffer, float value)
+{
+    return bytecode_put_bytes(buffer, &value, sizeof(value));
+}
+
+static inline size_t put_string(struct vkd3d_bytecode_buffer *buffer, const char *string)
+{
+    return bytecode_put_bytes(buffer, string, strlen(string) + 1);
+}
+
+static inline size_t bytecode_get_size(struct vkd3d_bytecode_buffer *buffer)
+{
+    return buffer->size;
+}
+
+uint32_t vkd3d_parse_integer(const char *s);
 
 struct vkd3d_shader_message_context
 {
     enum vkd3d_shader_log_level log_level;
-    const char *source_name;
-    unsigned int line, column;
     struct vkd3d_string_buffer messages;
 };
 
-void vkd3d_shader_message_context_cleanup(struct vkd3d_shader_message_context *context) DECLSPEC_HIDDEN;
-char *vkd3d_shader_message_context_copy_messages(struct vkd3d_shader_message_context *context) DECLSPEC_HIDDEN;
-bool vkd3d_shader_message_context_init(struct vkd3d_shader_message_context *context,
-        enum vkd3d_shader_log_level log_level, const char *source_name) DECLSPEC_HIDDEN;
+void vkd3d_shader_message_context_cleanup(struct vkd3d_shader_message_context *context);
+bool vkd3d_shader_message_context_copy_messages(struct vkd3d_shader_message_context *context, char **out);
+void vkd3d_shader_message_context_init(struct vkd3d_shader_message_context *context,
+        enum vkd3d_shader_log_level log_level);
 void vkd3d_shader_message_context_trace_messages_(const struct vkd3d_shader_message_context *context,
-        const char *function) DECLSPEC_HIDDEN;
+        const char *function);
 #define vkd3d_shader_message_context_trace_messages(context) \
         vkd3d_shader_message_context_trace_messages_(context, __FUNCTION__)
-void vkd3d_shader_error(struct vkd3d_shader_message_context *context, enum vkd3d_shader_error error,
-        const char *format, ...) VKD3D_PRINTF_FUNC(3, 4) DECLSPEC_HIDDEN;
-void vkd3d_shader_verror(struct vkd3d_shader_message_context *context,
-        enum vkd3d_shader_error error, const char *format, va_list args) DECLSPEC_HIDDEN;
+void vkd3d_shader_error(struct vkd3d_shader_message_context *context, const struct vkd3d_shader_location *location,
+        enum vkd3d_shader_error error, const char *format, ...) VKD3D_PRINTF_FUNC(4, 5);
+void vkd3d_shader_verror(struct vkd3d_shader_message_context *context, const struct vkd3d_shader_location *location,
+        enum vkd3d_shader_error error, const char *format, va_list args);
+void vkd3d_shader_vnote(struct vkd3d_shader_message_context *context, const struct vkd3d_shader_location *location,
+        enum vkd3d_shader_log_level level, const char *format, va_list args);
+void vkd3d_shader_vwarning(struct vkd3d_shader_message_context *context, const struct vkd3d_shader_location *location,
+        enum vkd3d_shader_error error, const char *format, va_list args);
 
-int shader_extract_from_dxbc(const void *dxbc, size_t dxbc_length,
-        struct vkd3d_shader_message_context *message_context, struct vkd3d_shader_desc *desc) DECLSPEC_HIDDEN;
-void free_shader_desc(struct vkd3d_shader_desc *desc) DECLSPEC_HIDDEN;
+void vkd3d_shader_dump_shader(enum vkd3d_shader_source_type source_type,
+        enum vkd3d_shader_type shader_type, const struct vkd3d_shader_code *shader);
+void vkd3d_shader_trace_text_(const char *text, size_t size, const char *function);
+#define vkd3d_shader_trace_text(text, size) \
+        vkd3d_shader_trace_text_(text, size, __FUNCTION__)
+
+int vkd3d_shader_sm1_parser_create(const struct vkd3d_shader_compile_info *compile_info,
+        struct vkd3d_shader_message_context *message_context, struct vkd3d_shader_parser **parser);
+int vkd3d_shader_sm4_parser_create(const struct vkd3d_shader_compile_info *compile_info,
+        struct vkd3d_shader_message_context *message_context, struct vkd3d_shader_parser **parser);
+
+void free_shader_desc(struct vkd3d_shader_desc *desc);
 
 int shader_parse_input_signature(const void *dxbc, size_t dxbc_length,
-        struct vkd3d_shader_message_context *message_context, struct vkd3d_shader_signature *signature) DECLSPEC_HIDDEN;
+        struct vkd3d_shader_message_context *message_context, struct vkd3d_shader_signature *signature);
+
+struct vkd3d_glsl_generator;
+
+struct vkd3d_glsl_generator *vkd3d_glsl_generator_create(const struct vkd3d_shader_version *version,
+        struct vkd3d_shader_message_context *message_context, const struct vkd3d_shader_location *location);
+int vkd3d_glsl_generator_generate(struct vkd3d_glsl_generator *generator,
+        struct vkd3d_shader_parser *parser, struct vkd3d_shader_code *out);
+void vkd3d_glsl_generator_destroy(struct vkd3d_glsl_generator *generator);
 
 struct vkd3d_dxbc_compiler;
 
 struct vkd3d_dxbc_compiler *vkd3d_dxbc_compiler_create(const struct vkd3d_shader_version *shader_version,
         const struct vkd3d_shader_desc *shader_desc, const struct vkd3d_shader_compile_info *compile_info,
         const struct vkd3d_shader_scan_descriptor_info *scan_descriptor_info,
-        struct vkd3d_shader_message_context *message_context) DECLSPEC_HIDDEN;
+        struct vkd3d_shader_message_context *message_context, const struct vkd3d_shader_location *location);
 int vkd3d_dxbc_compiler_handle_instruction(struct vkd3d_dxbc_compiler *compiler,
-        const struct vkd3d_shader_instruction *instruction) DECLSPEC_HIDDEN;
+        const struct vkd3d_shader_instruction *instruction);
 int vkd3d_dxbc_compiler_generate_spirv(struct vkd3d_dxbc_compiler *compiler,
-        const struct vkd3d_shader_compile_info *compile_info, struct vkd3d_shader_code *spirv) DECLSPEC_HIDDEN;
-void vkd3d_dxbc_compiler_destroy(struct vkd3d_dxbc_compiler *compiler) DECLSPEC_HIDDEN;
+        const struct vkd3d_shader_compile_info *compile_info, struct vkd3d_shader_code *spirv);
+void vkd3d_dxbc_compiler_destroy(struct vkd3d_dxbc_compiler *compiler);
 
-void vkd3d_compute_dxbc_checksum(const void *dxbc, size_t size, uint32_t checksum[4]) DECLSPEC_HIDDEN;
+void vkd3d_compute_dxbc_checksum(const void *dxbc, size_t size, uint32_t checksum[4]);
+
+int preproc_lexer_parse(const struct vkd3d_shader_compile_info *compile_info,
+        struct vkd3d_shader_code *out, struct vkd3d_shader_message_context *message_context);
+
+int hlsl_compile_shader(const struct vkd3d_shader_code *hlsl, const struct vkd3d_shader_compile_info *compile_info,
+        struct vkd3d_shader_code *out, struct vkd3d_shader_message_context *message_context);
 
 static inline enum vkd3d_shader_component_type vkd3d_component_type_from_data_type(
         enum vkd3d_data_type data_type)
@@ -889,8 +1118,12 @@ static inline enum vkd3d_shader_component_type vkd3d_component_type_from_data_ty
             return VKD3D_SHADER_COMPONENT_UINT;
         case VKD3D_DATA_INT:
             return VKD3D_SHADER_COMPONENT_INT;
+        case VKD3D_DATA_DOUBLE:
+            return VKD3D_SHADER_COMPONENT_DOUBLE;
         default:
             FIXME("Unhandled data type %#x.\n", data_type);
+            /* fall-through */
+        case VKD3D_DATA_MIXED:
             return VKD3D_SHADER_COMPONENT_UINT;
     }
 }
@@ -906,6 +1139,8 @@ static inline enum vkd3d_data_type vkd3d_data_type_from_component_type(
             return VKD3D_DATA_UINT;
         case VKD3D_SHADER_COMPONENT_INT:
             return VKD3D_DATA_INT;
+        case VKD3D_SHADER_COMPONENT_DOUBLE:
+            return VKD3D_DATA_DOUBLE;
         default:
             FIXME("Unhandled component type %#x.\n", component_type);
             return VKD3D_DATA_FLOAT;
@@ -940,12 +1175,31 @@ static inline unsigned int vkd3d_write_mask_from_component_count(unsigned int co
     return (VKD3DSP_WRITEMASK_0 << component_count) - 1;
 }
 
+static inline unsigned int vkd3d_write_mask_64_from_32(DWORD write_mask32)
+{
+    unsigned int write_mask64 = write_mask32 | (write_mask32 >> 1);
+    return (write_mask64 & VKD3DSP_WRITEMASK_0) | ((write_mask64 & VKD3DSP_WRITEMASK_2) >> 1);
+}
+
+static inline unsigned int vkd3d_write_mask_32_from_64(unsigned int write_mask64)
+{
+    unsigned int write_mask32 = (write_mask64 | (write_mask64 << 1))
+            & (VKD3DSP_WRITEMASK_0 | VKD3DSP_WRITEMASK_2);
+    return write_mask32 | (write_mask32 << 1);
+}
+
 static inline unsigned int vkd3d_swizzle_get_component(DWORD swizzle,
         unsigned int idx)
 {
     return (swizzle >> VKD3D_SHADER_SWIZZLE_SHIFT(idx)) & VKD3D_SHADER_SWIZZLE_MASK;
 }
 
+static inline unsigned int vkd3d_swizzle_get_component64(DWORD swizzle,
+        unsigned int idx)
+{
+    return ((swizzle >> VKD3D_SHADER_SWIZZLE_SHIFT(idx * 2)) & VKD3D_SHADER_SWIZZLE_MASK) / 2u;
+}
+
 static inline unsigned int vkd3d_compact_swizzle(unsigned int swizzle, unsigned int write_mask)
 {
     unsigned int i, compacted_swizzle = 0;
@@ -986,4 +1240,40 @@ static inline void *vkd3d_find_struct_(const struct vkd3d_struct *chain,
 #define VKD3D_DXBC_MAX_SOURCE_COUNT 6
 #define VKD3D_DXBC_HEADER_SIZE (8 * sizeof(uint32_t))
 
+#define TAG_AON9 VKD3D_MAKE_TAG('A', 'o', 'n', '9')
+#define TAG_DXBC VKD3D_MAKE_TAG('D', 'X', 'B', 'C')
+#define TAG_DXIL VKD3D_MAKE_TAG('D', 'X', 'I', 'L')
+#define TAG_ISG1 VKD3D_MAKE_TAG('I', 'S', 'G', '1')
+#define TAG_ISGN VKD3D_MAKE_TAG('I', 'S', 'G', 'N')
+#define TAG_OSG1 VKD3D_MAKE_TAG('O', 'S', 'G', '1')
+#define TAG_OSG5 VKD3D_MAKE_TAG('O', 'S', 'G', '5')
+#define TAG_OSGN VKD3D_MAKE_TAG('O', 'S', 'G', 'N')
+#define TAG_PCSG VKD3D_MAKE_TAG('P', 'C', 'S', 'G')
+#define TAG_PSG1 VKD3D_MAKE_TAG('P', 'S', 'G', '1')
+#define TAG_RD11 VKD3D_MAKE_TAG('R', 'D', '1', '1')
+#define TAG_RDEF VKD3D_MAKE_TAG('R', 'D', 'E', 'F')
+#define TAG_RTS0 VKD3D_MAKE_TAG('R', 'T', 'S', '0')
+#define TAG_SHDR VKD3D_MAKE_TAG('S', 'H', 'D', 'R')
+#define TAG_SHEX VKD3D_MAKE_TAG('S', 'H', 'E', 'X')
+#define TAG_TEXT VKD3D_MAKE_TAG('T', 'E', 'X', 'T')
+
+struct dxbc_writer_section
+{
+    uint32_t tag;
+    const uint8_t *data;
+    size_t size;
+};
+
+#define DXBC_MAX_SECTION_COUNT 5
+
+struct dxbc_writer
+{
+    unsigned int section_count;
+    struct dxbc_writer_section sections[DXBC_MAX_SECTION_COUNT];
+};
+
+void dxbc_writer_add_section(struct dxbc_writer *dxbc, uint32_t tag, const void *data, size_t size);
+void dxbc_writer_init(struct dxbc_writer *dxbc);
+int dxbc_writer_write(struct dxbc_writer *dxbc, struct vkd3d_shader_code *code);
+
 #endif  /* __VKD3D_SHADER_PRIVATE_H */
