@@ -17,9 +17,11 @@
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
  */
 
+#ifndef _GNU_SOURCE
 #define _GNU_SOURCE
+#endif
+#include <assert.h>
 #include <fcntl.h>
-#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <sys/stat.h>
@@ -28,32 +30,49 @@
 
 #include "vkd3d_common.h"
 #include "vkd3d_shader.h"
+#ifdef HAVE_NCURSES
+#include <term.h>
+#endif
 
-#define MAX_COMPILE_OPTIONS 3
+#define MAX_COMPILE_OPTIONS 4
 
 enum
 {
     OPTION_HELP = CHAR_MAX + 1,
     OPTION_BUFFER_UAV,
+    OPTION_ENTRY,
     OPTION_OUTPUT,
     OPTION_PRINT_SOURCE_TYPES,
     OPTION_PRINT_TARGET_TYPES,
+    OPTION_PROFILE,
     OPTION_STRIP_DEBUG,
     OPTION_VERSION,
     OPTION_TEXT_FORMATTING,
 };
 
+#define TAG_DXBC VKD3D_MAKE_TAG('D', 'X', 'B', 'C')
+
 static const struct source_type_info
 {
     enum vkd3d_shader_source_type type;
     const char *name;
     const char *description;
+    bool is_binary;
+    enum vkd3d_shader_target_type default_target_type;
 }
 source_type_info[] =
 {
     {VKD3D_SHADER_SOURCE_DXBC_TPF,
         "dxbc-tpf",     "A 'Tokenized Program Format' shader embedded in a DXBC container.\n"
-        "                This is the format used for Direct3D shader model 4 and 5 shaders.\n"},
+        "                This is the format used for Direct3D shader model 4 and 5 shaders.\n",
+        true, VKD3D_SHADER_TARGET_SPIRV_BINARY},
+    {VKD3D_SHADER_SOURCE_HLSL,
+        "hlsl",         "High Level Shader Language source code.\n",
+        false, VKD3D_SHADER_TARGET_DXBC_TPF},
+    {VKD3D_SHADER_SOURCE_D3D_BYTECODE,
+        "d3dbc",        "Legacy Direct3D byte-code.\n"
+        "                This is the format used for Direct3D shader model 1, 2, and 3 shaders.\n",
+        true, VKD3D_SHADER_TARGET_SPIRV_BINARY},
 };
 
 static const struct target_type_info
@@ -72,6 +91,20 @@ target_type_info[] =
     {VKD3D_SHADER_TARGET_SPIRV_TEXT,
         "spirv-text", "A SPIR-V shader in text form.\n",
         false},
+    {VKD3D_SHADER_TARGET_D3D_ASM,
+        "d3d-asm", "A shader in Direct3D assembly form.\n",
+        false},
+    {VKD3D_SHADER_TARGET_D3D_BYTECODE,
+        "d3dbc",        "Legacy Direct3D byte-code.\n"
+        "                This is the format used for Direct3D shader model 1, 2, and 3 shaders.\n",
+        true},
+    {VKD3D_SHADER_TARGET_DXBC_TPF,
+        "dxbc-tpf",     "A 'Tokenized Program Format' shader embedded in a DXBC container.\n"
+        "                This is the format used for Direct3D shader model 4 and 5 shaders.\n",
+        true},
+    {VKD3D_SHADER_TARGET_GLSL,
+        "glsl", "An 'OpenGL Shading Language' shader.\n",
+        false}
 };
 
 static bool read_shader(struct vkd3d_shader_code *shader, FILE *f)
@@ -140,10 +173,14 @@ static void print_usage(const char *program_name)
         "[options...] [file]\n"
         "Options:\n"
         "  -h, --help            Display this information and exit.\n"
-        "  -b <type>             Specify the target type.\n"
+        "  -b <type>             Specify the target type. Use --print-target-types to\n"
+        "                        list the valid and default target types for a given\n"
+        "                        source type.\n"
         "  --buffer-uav=<type>   Specify the buffer type to use for buffer UAV bindings.\n"
         "                        Valid values are 'buffer-texture' (default) and\n"
         "                        'storage-buffer'.\n"
+        "  -e, --entry=<name>    Use <name> as the entry point (default is \"main\").\n"
+        "  -E                    Preprocess the source code instead of compiling it.\n"
         "  -o, --output=<file>   Write the output to <file>. If <file> is '-' or no\n"
         "                        output file is specified, output will be written to\n"
         "                        standard output.\n"
@@ -155,10 +192,12 @@ static void print_usage(const char *program_name)
         "  --print-source-types  Display the supported source types and exit.\n"
         "  --print-target-types  Display the supported target types for the specified\n"
         "                        source type and exit.\n"
+        "  -p, --profile=<name>  Specify the target shader profile for HLSL shaders.\n"
         "  --strip-debug         Strip debug information from the output.\n"
         "  -V, --version         Display version information and exit.\n"
-        "  -x <type>             Specify the type of the source. Valid values are\n"
-        "                        'dxbc-tpf' and 'none'.\n"
+        "  -x <type>             Specify the type of the source. Use --print-source-types\n"
+        "                        to list valid source types. The default source type is\n"
+        "                        automatically detected from the input shader.\n"
         "  --                    Stop option processing. Any subsequent argument is\n"
         "                        interpreted as a filename.\n"
         "\n"
@@ -172,11 +211,17 @@ struct options
 {
     const char *filename;
     const char *output_filename;
+    const char *entry_point;
+    const char *profile;
     enum vkd3d_shader_source_type source_type;
     enum vkd3d_shader_target_type target_type;
+    bool preprocess_only;
+    bool print_help;
     bool print_version;
     bool print_source_types;
     bool print_target_types;
+    uint32_t formatting;
+    bool explicit_colour;
 
     struct vkd3d_shader_compile_option compile_options[MAX_COMPILE_OPTIONS];
     unsigned int compile_option_count;
@@ -227,7 +272,7 @@ static bool parse_buffer_uav(enum vkd3d_shader_compile_option_buffer_uav *buffer
     return false;
 }
 
-static bool parse_formatting(uint32_t *formatting, char *arg)
+static bool parse_formatting(uint32_t *formatting, bool *colour, char *arg)
 {
     static const struct formatting_option
     {
@@ -267,6 +312,8 @@ static bool parse_formatting(uint32_t *formatting, char *arg)
                     *formatting |= opts[i].value;
                 else
                     *formatting &= ~opts[i].value;
+                if (opts[i].value == VKD3D_SHADER_COMPILE_OPTION_FORMATTING_COLOUR)
+                    *colour = true;
                 break;
             }
         }
@@ -339,7 +386,7 @@ static bool validate_target_type(
     const enum vkd3d_shader_target_type *supported_types =
         vkd3d_shader_get_supported_target_types(source_type, &count);
 
-    for (i = 0; i < ARRAY_SIZE(target_type_info); ++i)
+    for (i = 0; i < count; ++i)
     {
         if (target_type == supported_types[i])
             return true;
@@ -350,8 +397,6 @@ static bool validate_target_type(
 
 static bool parse_command_line(int argc, char **argv, struct options *options)
 {
-    enum vkd3d_shader_compile_option_formatting_flags formatting = VKD3D_SHADER_COMPILE_OPTION_FORMATTING_INDENT
-            | VKD3D_SHADER_COMPILE_OPTION_FORMATTING_HEADER;
     enum vkd3d_shader_compile_option_buffer_uav buffer_uav;
     int option;
 
@@ -359,22 +404,26 @@ static bool parse_command_line(int argc, char **argv, struct options *options)
     {
         {"help",               no_argument,       NULL, OPTION_HELP},
         {"buffer-uav",         required_argument, NULL, OPTION_BUFFER_UAV},
+        {"entry",              required_argument, NULL, OPTION_ENTRY},
         {"output",             required_argument, NULL, OPTION_OUTPUT},
         {"formatting",         required_argument, NULL, OPTION_TEXT_FORMATTING},
         {"print-source-types", no_argument,       NULL, OPTION_PRINT_SOURCE_TYPES},
         {"print-target-types", no_argument,       NULL, OPTION_PRINT_TARGET_TYPES},
+        {"profile",            required_argument, NULL, OPTION_PROFILE},
         {"strip-debug",        no_argument,       NULL, OPTION_STRIP_DEBUG},
         {"version",            no_argument,       NULL, OPTION_VERSION},
         {NULL,                 0,                 NULL, 0},
     };
 
     memset(options, 0, sizeof(*options));
-    options->source_type = VKD3D_SHADER_SOURCE_DXBC_TPF;
-    options->target_type = VKD3D_SHADER_TARGET_SPIRV_BINARY;
+    options->source_type = VKD3D_SHADER_SOURCE_NONE;
+    options->target_type = VKD3D_SHADER_TARGET_NONE;
+    options->formatting = VKD3D_SHADER_COMPILE_OPTION_FORMATTING_INDENT
+            | VKD3D_SHADER_COMPILE_OPTION_FORMATTING_HEADER;
 
     for (;;)
     {
-        if ((option = getopt_long(argc, argv, "b:ho:Vx:", long_options, NULL)) == -1)
+        if ((option = getopt_long(argc, argv, "b:e:Eho:p:Vx:", long_options, NULL)) == -1)
             break;
 
         switch (option)
@@ -396,13 +445,32 @@ static bool parse_command_line(int argc, char **argv, struct options *options)
                 add_compile_option(options, VKD3D_SHADER_COMPILE_OPTION_BUFFER_UAV, buffer_uav);
                 break;
 
+            case OPTION_ENTRY:
+            case 'e':
+                options->entry_point = optarg;
+                break;
+
+            case 'E':
+                options->preprocess_only = true;
+                break;
+
+            case 'h':
+            case OPTION_HELP:
+                options->print_help = true;
+                return true;
+
             case OPTION_OUTPUT:
             case 'o':
                 options->output_filename = optarg;
                 break;
 
+            case OPTION_PROFILE:
+            case 'p':
+                options->profile = optarg;
+                break;
+
             case OPTION_TEXT_FORMATTING:
-                if (!parse_formatting(&formatting, optarg))
+                if (!parse_formatting(&options->formatting, &options->explicit_colour, optarg))
                     return false;
                 break;
 
@@ -436,18 +504,6 @@ static bool parse_command_line(int argc, char **argv, struct options *options)
         }
     }
 
-    if (!validate_target_type(options->source_type, options->target_type))
-    {
-        fprintf(stderr, "Target type '%s' is invalid for source type '%s'.\n",
-                get_target_type_info(options->target_type)->name,
-                get_source_type_info(options->source_type)->name);
-        return false;
-    }
-
-    if (options->print_target_types)
-        return true;
-
-    add_compile_option(options, VKD3D_SHADER_COMPILE_OPTION_FORMATTING, formatting);
     if (optind < argc)
         options->filename = argv[argc - 1];
 
@@ -472,8 +528,9 @@ static void print_source_types(void)
 
 static void print_target_types(enum vkd3d_shader_source_type source_type)
 {
+    const struct source_type_info *source_info = get_source_type_info(source_type);
     const enum vkd3d_shader_target_type *target_types;
-    const char *source_type_name = get_source_type_info(source_type)->name;
+    const char *source_type_name = source_info->name;
     unsigned int count, i;
 
     target_types = vkd3d_shader_get_supported_target_types(source_type, &count);
@@ -484,6 +541,7 @@ static void print_target_types(enum vkd3d_shader_source_type source_type)
         if (type)
             fprintf(stdout, "  %-12s  %s", type->name, type->description);
     }
+    printf("The default target type is '%s'.\n", get_target_type_info(source_info->default_target_type)->name);
 }
 
 static FILE *open_input(const char *filename, bool *close)
@@ -524,11 +582,33 @@ static FILE *open_output(const char *filename, bool *close)
     return f;
 }
 
+static bool has_colour(FILE *f)
+{
+#ifdef HAVE_NCURSES
+    bool supported;
+    int ret;
+
+    if (!isatty(fileno(f)))
+        return false;
+    setupterm(NULL, fileno(f), &ret);
+    if (ret != 1)
+        return false;
+    supported = !!tigetstr("setaf");
+    del_curterm(cur_term);
+
+    return supported;
+#else
+    return false;
+#endif
+}
+
 int main(int argc, char **argv)
 {
+    struct vkd3d_shader_spirv_target_info spirv_target_info = {0};
+    struct vkd3d_shader_hlsl_source_info hlsl_source_info = {0};
     bool close_input = false, close_output = false;
     struct vkd3d_shader_compile_info info;
-    struct vkd3d_shader_code spirv;
+    struct vkd3d_shader_code output_code;
     struct options options;
     FILE *input, *output;
     char *messages;
@@ -536,9 +616,12 @@ int main(int argc, char **argv)
     int ret;
 
     if (!parse_command_line(argc, argv, &options))
+        return 1;
+
+    if (options.print_help)
     {
         print_usage(argv[0]);
-        return 1;
+        return 0;
     }
 
     if (options.print_version)
@@ -557,6 +640,12 @@ int main(int argc, char **argv)
 
     if (options.print_target_types)
     {
+        if (options.source_type == VKD3D_SHADER_SOURCE_NONE)
+        {
+            fprintf(stderr, "--print-target-types requires an explicitly specified source type.\n");
+            return 1;
+        }
+
         print_target_types(options.source_type);
         return 0;
     }
@@ -564,7 +653,50 @@ int main(int argc, char **argv)
     if (!(input = open_input(options.filename, &close_input)))
         goto done;
 
-    if (!options.filename && isatty(fileno(input)))
+    if (!read_shader(&info.source, input))
+    {
+        fprintf(stderr, "Failed to read input shader.\n");
+        goto done;
+    }
+
+    if (options.source_type == VKD3D_SHADER_SOURCE_NONE)
+    {
+        uint32_t token;
+
+        if (options.preprocess_only || info.source.size < sizeof(token))
+        {
+            options.source_type = VKD3D_SHADER_SOURCE_HLSL;
+        }
+        else
+        {
+            memcpy(&token, info.source.code, sizeof(token));
+            if (token == TAG_DXBC)
+                options.source_type = VKD3D_SHADER_SOURCE_DXBC_TPF;
+            else if ((token & 0xfffe0000) == 0xfffe0000)
+                options.source_type = VKD3D_SHADER_SOURCE_D3D_BYTECODE;
+            else
+                options.source_type = VKD3D_SHADER_SOURCE_HLSL;
+        }
+    }
+
+    if (options.target_type == VKD3D_SHADER_TARGET_NONE && !options.preprocess_only)
+        options.target_type = get_source_type_info(options.source_type)->default_target_type;
+
+    if (!options.preprocess_only && !validate_target_type(options.source_type, options.target_type))
+    {
+        fprintf(stderr, "Target type '%s' is invalid for source type '%s'.\n",
+                get_target_type_info(options.target_type)->name,
+                get_source_type_info(options.source_type)->name);
+        return 1;
+    }
+
+    if (!options.preprocess_only && options.source_type == VKD3D_SHADER_SOURCE_HLSL && !options.profile)
+    {
+        fprintf(stderr, "You need to specify a profile when compiling from HLSL source.\n");
+        return 1;
+    }
+
+    if (!options.filename && get_source_type_info(options.source_type)->is_binary && isatty(fileno(input)))
     {
         fprintf(stderr, "Input is a tty and input format is binary, exiting.\n"
                 "If this is really what you intended, specify the input explicitly.\n");
@@ -574,16 +706,30 @@ int main(int argc, char **argv)
     if (!(output = open_output(options.output_filename, &close_output)))
         goto done;
 
-    if (!options.output_filename && get_target_type_info(options.target_type)->is_binary
-            && isatty(fileno(output)))
+    if (!options.output_filename && isatty(fileno(output)))
     {
-        fprintf(stderr, "Output is a tty and output format is binary, exiting.\n"
-                "If this is really what you intended, specify the output explicitly.\n");
-        goto done;
+        bool is_binary;
+
+        if (options.preprocess_only)
+            is_binary = get_source_type_info(options.source_type)->is_binary;
+        else
+            is_binary = get_target_type_info(options.target_type)->is_binary;
+
+        if (is_binary)
+        {
+            fprintf(stderr, "Output is a tty and output format is binary, exiting.\n"
+                    "If this is really what you intended, specify the output explicitly.\n");
+            goto done;
+        }
     }
 
+    if (!options.explicit_colour && !getenv("NO_COLOUR") && !getenv("NO_COLOR") && has_colour(output))
+        options.formatting |= VKD3D_SHADER_COMPILE_OPTION_FORMATTING_COLOUR;
+    add_compile_option(&options, VKD3D_SHADER_COMPILE_OPTION_FORMATTING, options.formatting);
+    add_compile_option(&options, VKD3D_SHADER_COMPILE_OPTION_API_VERSION, VKD3D_SHADER_API_VERSION_1_5);
+
     info.type = VKD3D_SHADER_STRUCTURE_TYPE_COMPILE_INFO;
-    info.next = NULL;
+    info.next = &hlsl_source_info;
     info.source_type = options.source_type;
     info.target_type = options.target_type;
     info.options = options.compile_options;
@@ -591,13 +737,20 @@ int main(int argc, char **argv)
     info.log_level = VKD3D_SHADER_LOG_INFO;
     info.source_name = options.filename;
 
-    if (!read_shader(&info.source, input))
-    {
-        fprintf(stderr, "Failed to read input shader.\n");
-        goto done;
-    }
+    hlsl_source_info.type = VKD3D_SHADER_STRUCTURE_TYPE_HLSL_SOURCE_INFO;
+    hlsl_source_info.next = &spirv_target_info;
+    hlsl_source_info.profile = options.profile;
+    hlsl_source_info.entry_point = options.entry_point;
+
+    spirv_target_info.type = VKD3D_SHADER_STRUCTURE_TYPE_SPIRV_TARGET_INFO;
+    spirv_target_info.entry_point = options.entry_point;
+    spirv_target_info.environment = VKD3D_SHADER_SPIRV_ENVIRONMENT_VULKAN_1_0;
+
+    if (options.preprocess_only)
+        ret = vkd3d_shader_preprocess(&info, &output_code, &messages);
+    else
+        ret = vkd3d_shader_compile(&info, &output_code, &messages);
 
-    ret = vkd3d_shader_compile(&info, &spirv, &messages);
     if (messages)
         fputs(messages, stderr);
     vkd3d_shader_free_messages(messages);
@@ -608,15 +761,15 @@ int main(int argc, char **argv)
         goto done;
     }
 
-    if (!write_shader(&spirv, output))
+    if (!write_shader(&output_code, output))
     {
         fprintf(stderr, "Failed to write output shader.\n");
-        vkd3d_shader_free_shader_code(&spirv);
+        vkd3d_shader_free_shader_code(&output_code);
         goto done;
     }
 
     fail = 0;
-    vkd3d_shader_free_shader_code(&spirv);
+    vkd3d_shader_free_shader_code(&output_code);
 done:
     if (close_output)
         fclose(output);
